"""
surepy.cli
====================================
The cli module of surepy

|license-info|
"""

from __future__ import annotations

import asyncio
import json

from datetime import datetime, time
from functools import wraps
from pathlib import Path
from shutil import copyfile
from sys import exit
from typing import Any, cast

import click

from aiohttp import ClientSession, TCPConnector
from rich import box
from rich.table import Table

from surepy import Surepy, __name__ as sp_name, __version__ as sp_version, console, natural_time
from surepy.entities.devices import Flap, SurepyDevice, Feeder
from surepy.entities.pet import Pet
from surepy.enums import Location, LockState


TOKEN_ENV = "SUREPY_TOKEN"


def coro(f: Any) -> Any:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(f(*args, **kwargs))

    return wrapper


token_file = Path("~/.surepy.token").expanduser()
old_token_file = token_file.with_suffix(".old_token")

CONTEXT_SETTINGS: dict[str, Any] = dict(help_option_names=["--help"])

version_message = (
    f" [#ffffff]{sp_name}[/] 🐾 [#666666]v[#aaaaaa]{sp_version.replace('.', '[#ff1d5e].[/]')}"
)


def print_header() -> None:
    """print header to terminal"""
    print()
    console.print(version_message, justify="left")
    print()


def token_available(ctx: click.Context) -> str | None:
    if token := ctx.obj.get("token"):
        return str(token)

    console.print("\n  [red bold]no token found![/]\n  checked in:\n")
    console.print("    · [bold]--token[/]")
    console.print(f"    · [bold]{TOKEN_ENV}[/] env var")
    console.print(f"    · [white bold]{token_file}[/]")
    console.print("\n\n  sorry 🐾 [bold]¯\\_(ツ)_/¯[/]\n\n")
    return None


# async def json_response(
#     data: dict[str, Any], ctx: click.Context, sp: Surepy | None = None
# ) -> None:
#     if ctx.obj.get("json", False):
#         if sp:
#             await sp.sac.close_session()
#
#         console.print(data)
#
#         exit(0)
#


@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.pass_context
@click.option("--version", default=False, is_flag=True, help=f"show {sp_name} version")
# @click.option("-v", "--verbose", default=False, is_flag=True, help="enable additional output")
# @click.option("-d", "--debug", default=False, is_flag=True, help="enable debug output")
@click.option("-j", "--json", default=False, is_flag=True, help="enable json api response output")
@click.option(
    "-t", "--token", "user_token", default=None, type=str, help="api token", hide_input=True
)
def cli(ctx: click.Context, json: bool, user_token: str, version: bool) -> None:
    """surepy cli 🐾

    https://github.com/benleb/surepy
    """

    ctx.ensure_object(dict)
    # ctx.obj["verbose"] = verbose
    # ctx.obj["debug"] = debug
    ctx.obj["json"] = json
    ctx.obj["token"] = user_token

    # if not json:
    #     print_header()

    if not ctx.invoked_subcommand:

        if version:
            click.echo(version_message)
            exit(0)

        click.echo(ctx.get_help())


@cli.command()
@click.pass_context
@click.option(
    "-u", "--user", required=True, type=str, help="sure petcare api account username (email)"
)
@click.option(
    "-p",
    "--password",
    required=True,
    type=str,
    help="sure petcare api account password",
    hide_input=True,
)
@coro
async def token(ctx: click.Context, user: str, password: str) -> None:
    """get a token"""

    surepy_token: str | None = None

    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        sp = Surepy(email=user, password=password, session=session)

        if surepy_token := await sp.sac.get_token():

            if token_file.exists() and surepy_token != token_file.read_text(encoding="utf-8"):
                copyfile(token_file, old_token_file)

            token_file.write_text(surepy_token, encoding="utf-8")

        # await sp.sac.close_session()

    console.rule(f"[bold]{user}[/] [#ff1d5e]·[/] [bold]Token[/]", style="#ff1d5e")
    console.print(f"[bold]{token}[/]", soft_wrap=True)
    console.rule(style="#ff1d5e")
    print()


@cli.command()
@click.pass_context
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def pets(ctx: click.Context, token: str | None) -> None:
    """get pets"""

    token = token if token else ctx.obj.get("token", None)

    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        sp = Surepy(auth_token=token, session=session)

        pets: list[Pet] = await sp.get_pets()

        # json output
        if ctx.obj.get("json", False):
            for pet in pets:
                json_str = json.dumps(pet.raw_data(), indent=4)
                print(json_str)
            return

        # pretty print
        table = Table(box=box.MINIMAL)
        table.add_column("Name", style="bold")
        table.add_column("Where", justify="right")
        table.add_column("Feeding A", justify="right", style="bold")
        table.add_column("Feeding B", justify="right", style="bold")
        table.add_column("Lunch Time", justify="right", style="bold")
        table.add_column("Drinking", justify="right", style="bold")
        table.add_column("Drink Time", justify="right", style="bold")
        table.add_column("ID 👤 ", justify="right")
        table.add_column("Household 🏡", justify="right")

        for pet in pets:

            feeding_a = feeding_b = lunch_time = None
            drinking_change = drink_time = None

            if pet.feeding:
                feeding_a = f"{pet.feeding.change[0]}g"
                feeding_b = f"{pet.feeding.change[1]}g"
                lunch_time = pet.feeding.at.time() if pet.feeding.at else None
            if pet.drinking:
                drinking_change = f"{pet.drinking.change[0]}ml"
                drink_time = pet.drinking.at.time() if pet.drinking.at else None

            table.add_row(
                str(pet.name),
                str(pet.location),
                f"{feeding_a}",
                f"{feeding_b}",
                str(lunch_time),
                f"{drinking_change}",
                str(drink_time),
                str(pet.pet_id),
                str(pet.household_id),
            )

        console.print(table, "", sep="\n")


@cli.command()
@click.pass_context
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def devices(ctx: click.Context, token: str | None) -> None:
    """get devices"""

    token = token if token else ctx.obj.get("token", None)

    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        sp = Surepy(auth_token=token, session=session)

        devices: list[SurepyDevice] = await sp.get_devices()

        # await json_response(devices, ctx)

        # table = Table(title="[bold][#ff1d5e]·[/] Devices [#ff1d5e]·[/]", box=box.MINIMAL)
        table = Table(box=box.MINIMAL)
        table.add_column("ID", justify="right", style="")
        table.add_column("Household", justify="right", style="")
        table.add_column("Name", style="bold")
        table.add_column("Type", style="")
        table.add_column("Serial", style="")

        # sorted_devices = sorted(devices, key=lambda x: int(devices[x]["household_id"]))

        # devices = await sp.sac.get_devices()
        # devices = await sp.get_entities()

        for device in devices:

            table.add_row(
                str(device.id),
                str(device.household_id),
                str(device.name),
                str(device.type.name.replace("_", " ").title()),
                str(device.serial) or "-",
            )

        console.print(table, "", sep="\n")


@cli.command()
@click.pass_context
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@click.option("-p", "--pet", "pet_id", required=False, type=int, help="id of the pet")
@click.option(
    "-h", "--household", "household_id", required=True, type=int, help="id of the household"
)
@coro
async def report(
    ctx: click.Context, household_id: int, pet_id: int | None = None, token: str | None = None
) -> None:
    """get pet/household report"""

    token = token if token else ctx.obj.get("token", None)

    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        sp = Surepy(auth_token=token, session=session)

        entities = await sp.get_entities()

        json_data = await sp.get_report(pet_id=pet_id, household_id=household_id)

        if data := json_data.get("data"):

            table = Table(box=box.MINIMAL)

            all_keys: list[str] = ["pet", "from", "to", "duration", "entry_device", "exit_device"]

            for key in all_keys:
                table.add_column(str(key))

            for pet in data:

                datapoints_drinking: list[dict[str, Any]] = pet.get("drinking", {}).get(
                    "datapoints", []
                )
                datapoints_feeding: list[dict[str, Any]] = pet.get("feeding", {}).get(
                    "datapoints", []
                )
                datapoints_movement: list[dict[str, Any]] = pet.get("movement", {}).get(
                    "datapoints", []
                )

                datapoints = datapoints_drinking + datapoints_feeding + datapoints_movement

                if datapoints:

                    for datapoint in datapoints:

                        from_time = datetime.fromisoformat(datapoint["from"])
                        to_time = (
                            datetime.fromisoformat(datapoint["to"])
                            if "active" not in datapoint
                            else None
                        )

                        if "active" in datapoint:
                            datapoint["duration"] = (
                                datetime.now(tz=from_time.tzinfo) - from_time
                            ).total_seconds()

                        entry_device = entities.get(datapoint.get("entry_device_id", 0), None)
                        exit_device = entities.pop(datapoint.get("exit_device_id", 0), None)

                        table.add_row(
                            str(entities[pet["pet_id"]].name),
                            str(from_time.strftime("%d/%m %H:%M")),
                            str(to_time.strftime("%d/%m %H:%M") if to_time else "-"),
                            str(natural_time(datapoint["duration"])),
                            str(entry_device.name if entry_device else "-"),
                            str(exit_device.name if exit_device else "-"),
                        )

            console.print(table, "", sep="\n")


@cli.command()
@click.pass_context
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def notification(ctx: click.Context, token: str | None = None) -> None:
    """get notifications"""

    token = token if token else ctx.obj.get("token", None)

    async with ClientSession(connector=TCPConnector(ssl=False)) as session:
        sp = Surepy(auth_token=token, session=session)

        json_data = await sp.get_notification() or None

        if json_data and (data := json_data.get("data")):

            table = Table(box=box.MINIMAL)

            all_keys: set[str] = set()
            all_keys.update(*[entry.keys() for entry in data])

            for key in all_keys:
                table.add_column(str(key))

            for entry in data:
                table.add_row(*([str(e) for e in entry.values()]))

            console.print(table, "", sep="\n")


@cli.command()
@click.pass_context
@click.option(
    "-d", "--device", "device_id", required=True, type=int, help="id of the sure petcare device"
)
@click.option(
    "-m",
    "--mode",
    required=True,
    type=click.Choice(["lock", "in", "out", "unlock"]),
    help="locking mode",
)
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def locking(ctx: click.Context, device_id: int, mode: str, token: str | None = None) -> None:
    """lock control"""

    token = token if token else ctx.obj.get("token", None)

    sp = Surepy(auth_token=str(token))

    if (flap := await sp.get_device(device_id=device_id)) and (type(flap) == Flap):

        flap = cast(Flap, flap)

        lock_state: LockState

        if mode == "lock":
            lock_state = LockState.LOCKED_ALL
            state = "locked"
        elif mode == "in":
            lock_state = LockState.LOCKED_IN
            state = "locked in"
        elif mode == "out":
            lock_state = LockState.LOCKED_OUT
            state = "locked out"
        elif mode == "unlock":
            lock_state = LockState.UNLOCKED
            state = "unlocked"
        else:
            return

        console.print(f"setting {flap.name} to '{state}'...")

        if await sp.sac._set_lock_state(device_id=device_id, mode=lock_state) and (
            device := await sp.get_device(device_id=device_id)
        ):
            console.print(f"✅ {device.name} set to '{state}' 🐾")
        else:
            console.print(f"❌ setting to '{state}' may have worked but something is fishy..!")

        # await sp.sac.close_session()


@cli.command()
@click.pass_context
@click.option(
    "-d", "--device", "device_id", required=True, type=int, help="id of the sure petcare device"
)
@click.option(
    "--lock-time",
    required=True,
    type=time.fromisoformat,
    help="Curfew lock time (in household's timezone)",
)
@click.option(
    "--unlock-time",
    required=True,
    type=time.fromisoformat,
    help="Curfew unlock time (in household's timezone)",
)
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def curfew(
    ctx: click.Context, device_id: int, lock_time: time, unlock_time: time, token: str | None = None
) -> None:
    """curfew control"""

    token = token if token else ctx.obj.get("token", None)

    sp = Surepy(auth_token=token)

    if (flap := await sp.get_device(device_id=device_id)) and (type(flap) == Flap):

        flap = cast(Flap, flap)

        console.print(
            f"setting {flap.name} curfew lock_time={str(lock_time)} unlock_time={str(unlock_time)}"
        )

        if await sp.sac.set_curfew(
            device_id=device_id, lock_time=lock_time, unlock_time=unlock_time
        ) and (device := await sp.get_device(device_id=device_id)):
            console.print(
                f"✅ {device.name} curfew lock_time={str(lock_time)} unlock_time={str(unlock_time)} 🐾"
            )
        else:
            console.print(
                f"❌ setting curfew lock_time={str(lock_time)} unlock_time={str(unlock_time)} may have worked but "
                f"something is fishy..!"
            )


@cli.command()
@click.pass_context
@click.option("--pet", "pet_id", required=True, type=int, help="id of the pet")
@click.option(
    "--position",
    required=True,
    type=click.Choice(["in", "out"]),
    help="position",
)
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def position(
    ctx: click.Context, pet_id: int, position: str, token: str | None = None
) -> None:
    """set pet position"""

    token = token if token else ctx.obj.get("token", None)

    sp = Surepy(auth_token=str(token))

    pet: Pet | None
    location: Location | None

    if (pet := await sp.get_pet(pet_id=pet_id)) and (type(pet) == Pet):

        if position == "in":
            location = Location.INSIDE
        elif position == "out":
            location = Location.OUTSIDE
        else:
            return

        if location:
            if await sp.sac.set_pet_location(pet.id, location):
                console.print(f"{pet.name} set to '{location.name}' 🐾")
            else:
                console.print(
                    f"setting to '{location.name}' probably worked but something else is fishy...!"
                )

        # await sp.sac.close_session()

@cli.command()
@click.pass_context
@click.option(
    "-d", "--device", "device_id", required=True, type=int, help="id of the sure petcare device"
)
@click.option("-p", "--pet", "pet_id", required=False, type=int, help="id of the pet")
@click.option(
    "-m",
    "--mode",
    required=True,
    type=click.Choice(["add", "remove", "list"]),
    help="assignment action",
)
@click.option(
    "-t", "--token", required=False, type=str, help="sure petcare api token", hide_input=True
)
@coro
async def feederassign(ctx: click.Context, device_id: int, mode: str, pet_id: int | None = None, token: str | None = None) -> None:
    """feeder pet assignment"""

    token = token if token else ctx.obj.get("token", None)

    sp = Surepy(auth_token=str(token))

    if (feeder := await sp.get_device(device_id=device_id)) and (type(feeder) == Feeder):

        pets: list[Pet] = await sp.get_pets()

        if mode == "list":
            table = Table(box=box.MINIMAL)
            table.add_column("ID", style="bold")
            table.add_column("Name", style="")
            table.add_column("Created At", style="")
            for tag in feeder.tags.values():
                for pet in pets:
                    if tag.id == pet.tag_id:
                        table.add_row(
                            str(pet.id),
                            str(pet.name),
                            str(datetime.fromisoformat(tag.created_at())),
                        )
            console.print(table, "", sep="\n")
        if mode == "add":
            for pet in pets:
                if pet.id == pet_id:
                    for tag in feeder.tags.values():
                        if tag.id == pet.tag_id:
                            console.print(f"Pet is already assigned to this feeder.")
                            return
                    if await sp.sac._add_tag_to_device(device_id=device_id, tag_id=pet.tag_id):
                        console.print(f"✅ {pet.name} added to '{feeder.name}' 🐾")
        if mode == "remove":
            for pet in pets:
                if pet.id == pet_id:
                    for tag in feeder.tags.values():
                        if tag.id == pet.tag_id:
                            if await sp.sac._remove_tag_from_device(device_id=device_id, tag_id=pet.tag_id):
                                console.print(f"✅ {pet.name} removed from '{feeder.name}' 🐾")
                                return
                    console.print("Pet is not assigned to this feeder.")
        else:
            return
        # await sp.sac.close_session()

if __name__ == "__main__":
    cli(obj={})



# battery voltages
SURE_BATT_VOLTAGE_FULL = 1.6
SURE_BATT_VOLTAGE_LOW = 1.2
SURE_BATT_VOLTAGE_DIFF = SURE_BATT_VOLTAGE_FULL - SURE_BATT_VOLTAGE_LOW

# HTTP user agent
SUREPY_USER_AGENT = "surepy {version} - https://github.com/benleb/surepy"

# Sure Petcare API endpoints
BASE_RESOURCE: str = "https://app.api.surehub.io/api"
AUTH_RESOURCE: str = f"{BASE_RESOURCE}/auth/login"
MESTART_RESOURCE: str = f"{BASE_RESOURCE}/me/start"
TIMELINE_RESOURCE: str = f"{BASE_RESOURCE}/timeline"
HOUSEHOLD_TIMELINE_RESOURCE: str = "{BASE_RESOURCE}/timeline/household/{household_id}?page={page}"
NOTIFICATION_RESOURCE: str = f"{BASE_RESOURCE}/notification"
PET_RESOURCE: str = f"{BASE_RESOURCE}/pet?with%5B%5D=photo&with%5B%5D=breed&with%5B%5D=conditions&with%5B%5D=tag&with%5B%5D=food_type&with%5B%5D=species&with%5B%5D=position&with%5B%5D=status"
DEVICE_RESOURCE: str = f"{BASE_RESOURCE}/device?with%5B%5D=children&with%5B%5D=tags&with%5B%5D=control&with%5B%5D=status"
CONTROL_RESOURCE: str = "{BASE_RESOURCE}/device/{device_id}/control"
POSITION_RESOURCE: str = "{BASE_RESOURCE}/pet/{pet_id}/position"
ATTRIBUTES_RESOURCE: str = f"{BASE_RESOURCE}/start"
DEVICE_TAG_RESOURCE: str = "{BASE_RESOURCE}/device/{device_id}/tag/{tag_id}"


API_TIMEOUT = 45

# HTTP constants
ACCEPT = "Accept"
ACCEPT_ENCODING = "Accept-Encoding"
ACCEPT_LANGUAGE = "Accept-Language"
AUTHORIZATION = "Authorization"
CONNECTION = "Connection"
CONTENT_TYPE_JSON = "application/json"
CONTENT_TYPE_TEXT_PLAIN = "text/plain"
ETAG = "Etag"
HOST = "Host"
HTTP_HEADER_X_REQUESTED_WITH = "X-Requested-With"
ORIGIN = "Origin"
REFERER = "Referer"
USER_AGENT = "User-Agent"



"""
surepy.entities.states
====================================
Classes representing pet states.

|license-info|
"""

from __future__ import annotations

from abc import ABC
from datetime import datetime
from typing import Any

from surepy.enums import Location


class PetState(ABC):
    """abstract surepy state."""

    def __init__(self, state: dict[str, dict[str, Any]]):
        self.activity: ActivityState | None = (
            ActivityState(state=state["activity"]) if "activity" in state else None
        )
        self.drinking: DrinkingState | None = (
            DrinkingState(state=state["drinking"]) if "drinking" in state else None
        )
        self.feeding: FeedingState | None = (
            FeedingState(state=state["feeding"]) if "feeding" in state else None
        )


class ActivityState:
    """surepy activity state."""

    def __init__(self, state: dict[str, Any]):
        self.device_id = state.get("device_id")
        self.tag_id = state.get("tag_id")
        self.since: datetime = (
            datetime.fromisoformat(state["at"]) if isinstance(state.get("at", None), str) else None
        )
        self.where: Location = Location(state["where"])


class DrinkingState:
    """surepy drinking state."""

    def __init__(self, state: dict[str, Any]):
        self.device_id = state.get("device_id")
        self.tag_id = state.get("tag_id")
        self.at: datetime = datetime.fromisoformat(state.get("at"))
        self.change: float = state["change"] if "change" in state else None


class FeedingState:
    """surepy feeding state."""

    def __init__(self, state: dict[str, Any]):
        self.device_id = state.get("device_id")
        self.tag_id = state.get("tag_id")
        self.at: datetime = datetime.fromisoformat(state.get("at"))
        self.changes: list[float] = state["change"] if "change" in state else None
        self.change_bowl_one = self.changes[0] if self.changes else None
        self.change_bowl_two = self.changes[1] if self.changes else None



"""
surepy
====================================
The core module of surepy

|license-info|
"""

from __future__ import annotations

import logging

from datetime import datetime
from importlib.metadata import version
from logging import Logger
from math import ceil
from typing import Any
from uuid import uuid1

import aiohttp

from rich.console import Console

from surepy.client import SureAPIClient, find_token, token_seems_valid
from surepy.const import (
    API_TIMEOUT,
    ATTRIBUTES_RESOURCE as ATTR_RESOURCE,
    BASE_RESOURCE,
    HOUSEHOLD_TIMELINE_RESOURCE,
    MESTART_RESOURCE,
    NOTIFICATION_RESOURCE,
    TIMELINE_RESOURCE,
)
from surepy.entities import SurepyEntity
from surepy.entities.devices import Feeder, Felaqua, Flap, Hub, SurepyDevice
from surepy.entities.pet import Pet
from surepy.enums import EntityType


__version__ = version(__name__)

# TOKEN_ENV = "SUREPY_TOKEN"  # nosec
# TOKEN_FILE = Path("~/.surepy.token").expanduser()

# get a logger
logger: Logger = logging.getLogger(__name__)

console = Console(width=120)


def natural_time(duration: int) -> str:
    """Transforms a number of seconds to a more human-friendly string.

    Args:
        duration (int): duration in seconds

    Returns:
        str: human-friendly duration string
    """

    duration_h, duration_min = divmod(duration, int(60 * 60))
    duration_min, duration_sec = divmod(duration_min, int(60))

    # append suitable unit
    if duration >= 60 * 60 * 24:
        duration_d, duration_h = divmod(duration_h, int(24))
        natural = f"{int(duration_d)}d {int(duration_h)}h {int(duration_min)}m"

    elif duration >= 60 * 60:
        if duration_min < 2 or duration_min > 58:
            natural = f"{int(duration_h)}h"
        else:
            natural = f"{int(duration_h)}h {int(duration_min)}m"

    elif duration > 60:
        natural = f"{int(duration_min)}min"

    else:
        natural = f"{int(duration_sec)}sec"

    return natural


class Surepy:
    """Communication with the Sure Petcare API."""

    def __init__(
        self,
        email: str | None = None,
        password: str | None = None,
        auth_token: str | None = None,
        api_timeout: int = API_TIMEOUT,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        """Initialize the connection to the Sure Petcare API."""

        # random device id
        self._device_id: str = str(uuid1())

        self._session = session

        self.sac = SureAPIClient(
            email=email,
            password=password,
            auth_token=auth_token,
            api_timeout=api_timeout,
            session=self._session,
            surepy_version=__version__,
        )

        # api token management
        self._auth_token: str | None = None
        if auth_token and token_seems_valid(auth_token):
            self._auth_token = auth_token
        else:  # if token := find_token():
            self._auth_token = find_token()

        self.entities: dict[int, SurepyEntity] = {}
        self._pets: dict[int, Any] = {}
        self._flaps: dict[int, Any] = {}
        self._feeders: dict[int, Any] = {}
        self._hubs: dict[int, Any] = {}

        self._breeds: dict[int, dict[int, Any]] = {}
        self._species_breeds: dict[int, dict[int, Any]] = {}
        self._conditions: dict[int, Any] = {}

        # storage for received api data
        self._resource: dict[str, Any] = {}
        # storage for etags
        self._etags: dict[str, str] = {}

        logger.debug("initialization completed | vars(): %s", vars())

    @property
    def auth_token(self) -> str | None:
        """Authentication token for device"""
        return self._auth_token

    async def pets_details(self) -> list[dict[str, Any]] | None:
        """Fetch pet information."""
        return await self.sac.get_pets()

    async def latest_actions(self, household_id: int) -> dict[int, dict[str, Any]] | None:
        """
        Args:
            household_id (int): ID associated with household
            pet_id (int): ID associated with pet

        Returns:
            Get the latest action using pet_id and household_id
            from raw data and output as a dictionary
        """
        return await self.get_actions(household_id=household_id)

    async def all_actions(self, household_id: int) -> dict[int, dict[str, Any]] | None:
        """Args:
        - household_id (int): id associated with household
        - pet_id (int): id associated with pet
        """
        return await self.get_actions(household_id=household_id)

    async def get_actions(self, household_id: int) -> dict[int, dict[str, Any]] | None:
        resource = f"{BASE_RESOURCE}/report/household/{household_id}"

        latest_actions: dict[int, dict[str, Any]] = {}

        pet_device_pairs: dict[str, Any] = (
            await self.sac.call(method="GET", resource=resource) or {}
        )

        if "data" not in pet_device_pairs:
            return latest_actions

        data: list[dict[str, Any]] = pet_device_pairs["data"]

        for pair in data:

            pet_id = int(pair["pet_id"])
            device_id = int(pair["device_id"])
            device: SurepyDevice = self.entities[device_id]  # type: ignore

            latest_actions[pet_id] = {}
            latest_actions[pet_id] = self.entities[device_id]._data

            # movement
            if (
                device.type in [EntityType.CAT_FLAP, EntityType.PET_FLAP]
                and pair["movement"]["datapoints"]
            ):
                latest_datapoint = pair["movement"]["datapoints"].pop()
                # latest_actions[pet_id]["move"] = latest_datapoint
                latest_actions[pet_id] = self.entities[device_id]._data["move"] = latest_datapoint

            # feeding
            elif (
                device.type in [EntityType.FEEDER, EntityType.FEEDER_LITE]
                and pair["feeding"]["datapoints"]
            ):
                latest_datapoint = pair["feeding"]["datapoints"].pop()
                # latest_actions[pet_id]["lunch"] = latest_datapoint
                latest_actions[pet_id] = self.entities[device_id]._data["lunch"] = latest_datapoint

            # drinking
            elif device.type == EntityType.FELAQUA and pair["drinking"]["datapoints"]:
                latest_datapoint = pair["drinking"]["datapoints"].pop()
                # latest_actions[pet_id]["drink"] = latest_datapoint
                latest_actions[pet_id] = self.entities[device_id]._data["drink"] = latest_datapoint

        return latest_actions

    async def get_latest_anonymous_drinks(self, household_id: int) -> dict[str, Any] | None:

        latest_drink: dict[str, float | str | datetime] = {}

        household_timeline = await self.get_household_timeline(household_id, entries=50)

        felaqua_related_entries: list[dict[str, Any]] = list(
            filter(
                lambda x: x["type"] in [29, 30, 34],  # type: ignore
                household_timeline,  # type: ignore
            )
        )

        if felaqua_related_entries:
            try:
                device_id = felaqua_related_entries[0]["weights"][0]["device_id"]
                latest_entry_frame = felaqua_related_entries[0]["weights"][0]["frames"][0]
                remaining = latest_entry_frame["current_weight"]
                change = latest_entry_frame["change"]
                updated_at = latest_entry_frame["updated_at"]
                latest_drink = {"remaining": remaining, "change": change, "date": updated_at}

                self.entities[device_id]._data["latest_drink"] = latest_drink

            except (KeyError, TypeError, IndexError):
                logger.warning(
                    "no water remaining/change events found in household timeline "
                    "(checked last %s entries)",
                    len(household_timeline) or 0,
                )

        return latest_drink

    async def get_household_timeline(
        self, household_id: int | None = None, entries: int = 25
    ) -> list[dict[str, Any]]:
        """Fetch Felaqua water level information."""

        # pagination as the api gives us at most 25 results per page
        max_entries_per_page = 25
        pages_to_fetch = ceil(entries / max_entries_per_page)

        current_page = 1
        household_timeline = []

        while current_page <= pages_to_fetch:

            resource = HOUSEHOLD_TIMELINE_RESOURCE.format(
                BASE_RESOURCE=BASE_RESOURCE,
                household_id=household_id,
                page=current_page,
                page_size=max_entries_per_page,
            )

            if timeline := await self.sac.call(method="GET", resource=resource):
                household_timeline += timeline.get("data", [])

            current_page += 1

        return household_timeline

    async def get_timeline(self) -> dict[str, Any]:
        """Retrieve the flap data/state."""
        return await self.sac.call(method="GET", resource=TIMELINE_RESOURCE) or {}

    async def get_notification(self) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        return await self.sac.call(
            method="GET", resource=NOTIFICATION_RESOURCE, timeout=API_TIMEOUT * 2
        )

    async def get_report(self, household_id: int, pet_id: int | None = None) -> dict[str, Any]:
        """Retrieve the pet/household report."""
        return (
            await self.sac.call(
                method="GET",
                resource=f"{BASE_RESOURCE}/report/household/{household_id}/pet/{pet_id}",
            )
            if pet_id
            else await self.sac.call(
                method="GET", resource=f"{BASE_RESOURCE}/report/household/{household_id}"
            )
        ) or {}

    async def get_pet(self, pet_id: int) -> Pet | None:
        if pet_id not in self.entities:
            await self.get_entities()

        if self.entities[pet_id].type == EntityType.PET:
            return self.entities[pet_id]  # type: ignore
        else:
            return None

    async def get_pets(self) -> list[Pet]:
        return [pet for pet in (await self.get_entities()).values() if isinstance(pet, Pet)]

    async def get_device(self, device_id: int) -> SurepyDevice | None:
        if device_id not in self.entities:
            await self.get_entities()

        if self.entities[device_id].type != EntityType.PET:
            return self.entities[device_id]  # type: ignore
        else:
            return None

    async def get_devices(self) -> list[SurepyDevice]:
        return [
            device
            for device in (await self.get_entities()).values()
            if isinstance(device, SurepyDevice)
        ]

    async def get_attributes(self) -> dict[str, Any] | None:
        # fetch additional data from sure petcare
        attributes: dict[str, Any] | None = None

        if (raw_data := (await self.sac.call(method="GET", resource=ATTR_RESOURCE))) and (
            attributes := raw_data.get("data")
        ):

            for breed in attributes.get("breed", {}):
                self._breeds[breed["id"]] = breed["name"]

                if breed["species_id"] not in self._breeds:
                    self._species_breeds[breed["species_id"]] = {}

                self._species_breeds[breed["species_id"]][breed["id"]] = breed["name"]

            for condition in attributes.get("condition", {}):
                self._conditions[condition["id"]] = condition["name"]

        return attributes

    async def get_entities(self, refresh: bool = False) -> dict[int, SurepyEntity]:
        """Get all Entities (Pets/Devices)"""

        household_ids: set[int] = set()
        felaqua_household_ids: set[int] = set()
        surepy_entities: dict[int, SurepyEntity] = {}

        raw_data: dict[str, list[dict[str, Any]]] = {}

        # get data like species, breed, conditions
        # await self.get_attributes()

        if MESTART_RESOURCE not in self.sac.resources or refresh:
            if response := await self.sac.call(method="GET", resource=MESTART_RESOURCE):
                raw_data = response.get("data", {})
        else:
            raw_data = self.sac.resources[MESTART_RESOURCE].get("data", {})

        if not raw_data:
            logger.error("could not fetch data ¯\\_(ツ)_/¯")
            return surepy_entities

        all_entities = raw_data.get("devices", []) + raw_data.get("pets", [])

        for entity in all_entities:

            # key used by sure petcare in api response
            entity_type = EntityType(int(entity.get("product_id", 0)))
            entity_id = entity["id"]

            if entity_type in [EntityType.CAT_FLAP, EntityType.PET_FLAP]:
                surepy_entities[entity_id] = Flap(data=entity)
            elif entity_type in [EntityType.FEEDER, EntityType.FEEDER_LITE]:
                surepy_entities[entity_id] = Feeder(data=entity)
            elif entity_type == EntityType.FELAQUA:
                surepy_entities[entity_id] = Felaqua(data=entity)
                felaqua_household_ids.add(int(surepy_entities[entity_id].household_id))
            elif entity_type == EntityType.HUB:
                surepy_entities[entity_id] = Hub(data=entity)
            elif entity_type == EntityType.PET:
                surepy_entities[entity_id] = Pet(data=entity)

            else:
                logger.warning(
                    "unknown type: %s (%s): %s", entity.get("name", "-"), entity_type, entity
                )

            household_ids.add(surepy_entities[entity_id].household_id)

            self.entities[entity_id] = surepy_entities[entity_id]

        # fetch additional data about movement, feeding & drinking
        for household_id in household_ids:
            await self.get_actions(household_id=household_id)
        for household_id in felaqua_household_ids:
            await self.get_latest_anonymous_drinks(household_id=household_id)

        # stupid idea, fix this
        _ = [
            feeder.add_bowls()  # type: ignore
            for feeder in surepy_entities.values()
            if feeder.type == EntityType.FEEDER
        ]

        return self.entities



"""
surepy.devices
====================================
ABC representing a Sure Petcare Device.

|license-info|
"""

from __future__ import annotations

import logging

from abc import ABC
from typing import Any
from urllib.parse import urlparse

from surepy.const import SURE_BATT_VOLTAGE_FULL, SURE_BATT_VOLTAGE_LOW
from surepy.entities import SurepyEntity
from surepy.enums import BowlPosition, FoodType, LockState


# get a logger
logger: logging.Logger = logging.getLogger(__name__)


class Hub(SurepyEntity):
    """Sure Petcare Hub."""

    @property
    def online(self) -> bool:
        return bool(self._data.get("status", {}).get("online"))

    @property
    def parent_id(self) -> int | None:
        return self._data.get("parent_device_id", None)

    @property
    def serial(self) -> str | None:
        """ID of the household the pet belongs to."""
        return str(serial) if (serial := self._data.get("serial_number")) else None

    @property
    def icon(self) -> str | None:
        """Picture of the Pet."""
        return urlparse("https://surehub.io/assets/images/hub-icon.svg").geturl()


class SurepyDevice(SurepyEntity, ABC):
    """Abstract Surepy base device"""

    @property
    def parent_id(self) -> int | None:
        return self._data.get("parent_device_id", None)

    @property
    def serial(self) -> str | None:
        """ID of the household the pet belongs to."""
        return str(serial) if (serial := self._data.get("serial_number")) else None

    @property
    def battery_level(self) -> int | None:
        """Return battery level in percent."""
        return self.calculate_battery_level()

    def calculate_battery_level(
        self,
        voltage_full: float = SURE_BATT_VOLTAGE_FULL,
        voltage_low: float = SURE_BATT_VOLTAGE_LOW,
        num_batteries: int = 4,
    ) -> int | None:
        """Return battery voltage."""

        try:
            voltage_diff = voltage_full - voltage_low
            battery_voltage = float(self._data["status"]["battery"])
            voltage_per_battery = battery_voltage / num_batteries
            voltage_per_battery_diff = voltage_per_battery - voltage_low

            # return batterie level between 0 and 100
            return max(min(int(voltage_per_battery_diff / voltage_diff * 100), 100), 0)

        except (KeyError, TypeError, ValueError) as error:
            logger.debug("error while calculating battery level: %s", error)
            return None


class FeederBowl:
    """Sure Petcare Felaqua."""

    def __init__(self, data: dict[str, int | float | str], feeder: Feeder):
        """Initialize a Sure Petcare sensor."""

        self._data: dict[str, int | float | str] = data
        self._name = f"{feeder.name} Bowl {self._data['index']}"

    @property
    def name(self) -> str:
        return self._name

    @property
    def weight(self) -> float:
        return float(self._data["weight"])

    @property
    def change(self) -> float:
        return float(self._data["change"])

    @property
    def target(self) -> int | None:
        return int(self._data["target"]) if "target" in self._data else None

    @property
    def index(self) -> int | None:
        return int(self._data["index"]) if "index" in self._data else None

    @property
    def food_type_id(self) -> int | None:
        return int(self._data["food_type_id"]) if "food_type_id" in self._data else None

    @property
    def food_type(self) -> str | None:
        return FoodType(self.food_type_id).name.capitalize() if self.food_type_id else None

    @property
    def position(self) -> str | None:
        return BowlPosition(self.index).name.capitalize() if self.index else None

    def raw_data(self) -> dict[str, int | float | str]:
        return self._data

class Tag:
    """Tags assigned to a device."""

    def __init__(self, data: dict[str, int | float | str], feeder: Feeder):
        """Initialize a Sure Petcare sensor."""

        self._data: dict[str, int | float | str] = data

    @property
    def id(self) -> int:
        return int(self._data["id"])

    def index(self) -> int:
        return int(self._data["index"])

    def profile(self) -> int:
        return int(self._data["profile"])

    def version(self) -> str:
        return self._data["version"]

    def created_at(self) -> str:
        return str(self._data["created_at"])

    def updated_at(self) -> str:
        return self._data["updated_at"]

    def raw_data(self) -> dict[str, int | float | str]:
        return self._data

class Feeder(SurepyDevice):
    """Sure Petcare Cat- or Pet-Flap."""

    def __init__(self, data: dict[str, Any]):
        """Initialize a Sure Petcare sensor."""
        super().__init__(data)

        self.bowls: dict[int, FeederBowl] = {}

        self.add_bowls()

        self.tags: dict[int, Tag] = {}

        self.add_tags()

    @property
    def bowl_count(self) -> int:
        return len(self.bowls)

    @property
    def total_weight(self) -> float:
        return sum([bowl.weight or 0.0 for bowl in self.bowls.values() if bowl.weight > 0.0])

    def add_bowls(self) -> None:
        if lunch := self._data.get("lunch"):
            for bowl in lunch.get("weights", []):
                self.bowls[bowl["index"]] = FeederBowl(data=bowl, feeder=self)

    @property
    def icon(self) -> str | None:
        """Icon of the Felaqua."""
        return urlparse("https://surehub.io/assets/images/feeder-left-menu.png").geturl()

    def add_tags(self) -> None:
        if tags := self._data.get("tags"):
            for tag in tags:
                self.tags[tag["index"]] = Tag(data=tag, feeder=self)

class Felaqua(SurepyDevice):
    """Sure Petcare Cat- or Pet-Flap."""

    @property
    def water_remaining(self) -> float | None:
        remaining = None

        try:
            remaining = float(self._data["latest_drink"]["remaining"])
        except (KeyError, TypeError):
            pass

        return remaining

    @property
    def water_change(self) -> float | None:
        change = None

        try:
            change = float(self._data["latest_drink"]["change"])
        except (KeyError, TypeError):
            pass

        return change

    @property
    def icon(self) -> str | None:
        """Icon of the Felaqua."""
        return urlparse("https://surehub.io/assets/images/poseidon-left-menu.png").geturl()


class Flap(SurepyDevice):
    """Sure Petcare Cat- or Pet-Flap."""

    @property
    def state(self) -> LockState:
        return LockState(self._data["status"]["locking"]["mode"])

    @property
    def unlocked(self) -> bool:
        return self.state in [LockState.UNLOCKED, LockState.CURFEW_UNLOCKED]

    @property
    def icon(self) -> str | None:
        """Icon of the Pet/Cap Flap."""

        icon_url = "https://surehub.io/assets/images/petdoor-left-menu.png"

        if self.state == LockState.LOCKED_ALL:
            icon_url = "https://surehub.io/assets/images/both-ways-icon.svg"
        elif self.state == LockState.LOCKED_IN:
            icon_url = "https://surehub.io/assets/images/inside-icon.svg"
        elif self.state == LockState.LOCKED_OUT:
            icon_url = "https://surehub.io/assets/images/outside-icon.svg"

        return urlparse(icon_url).geturl()



class SurePetcareError(Exception):
    """General Sure Petcare Error exception occurred."""


class SurePetcareConnectionError(SurePetcareError):
    """When a connection error is encountered."""


class SurePetcareAuthenticationError(SurePetcareError):
    """When a authentication error is encountered."""



from enum import IntEnum


class SureEnum(IntEnum):
    """Sure base enum."""

    def __str__(self) -> str:
        return self.name.title()  # pylint: disable=no-member


class EntityType(SureEnum):
    """Sure Entity Types."""

    PET = 0  # artificial ID, not used by the Sure Petcare API
    HUB = 1  # Hub
    REPEATER = 2  # Repeater
    PET_FLAP = 3  # Pet Door Connect
    FEEDER = 4  # Microchip Pet Feeder Connect
    PROGRAMMER = 5  # Programmer
    CAT_FLAP = 6  # Cat Flap Connect
    FEEDER_LITE = 7  # Feeder Lite
    FELAQUA = 8  # Felaqua Connect


class LockState(SureEnum):
    """Sure Petcare API State IDs."""

    UNLOCKED = 0
    LOCKED_IN = 1
    LOCKED_OUT = 2
    LOCKED_ALL = 3
    CURFEW = 4
    CURFEW_LOCKED = -1
    CURFEW_UNLOCKED = -2
    CURFEW_UNKNOWN = -3


class Location(SureEnum):
    """Sure Locations."""

    INSIDE = 1
    OUTSIDE = 2
    UNKNOWN = -1


class FoodType(SureEnum):
    """Sure Food Types."""

    WET = 1
    DRY = 2
    BOTH = 3
    UNKNOWN = -1


class BowlPosition(SureEnum):
    """Sure Feeder Bowl side."""

    LEFT = 0
    RIGHT = 1


class Species(SureEnum):
    """Species of the pet."""

    CAT = 1
    DOG = 2


class TimelineEvent(SureEnum):
    """Sure timeline event types."""

    BATTERY_LOW = 1
    FOOD_FILLED = 21
    EAT = 22
    FEEDER_TARE = 24
    DRINK = 29
    REMINDER_FRESH_WATER = 32
    ANONYMOUS_DRINK = 34



"""
surepy.pet
====================================
The `Pet` class of surepy

|license-info|
"""

from __future__ import annotations

from datetime import datetime
from typing import Any
from urllib.parse import urlparse

from surepy.entities import PetActivity, PetLocation, StateDrinking, StateFeeding, SurepyEntity
from surepy.entities.states import PetState
from surepy.enums import EntityType, FoodType, Location


class Pet(SurepyEntity):
    """
    Represents pet. Contains attributes of the pet.

    Attributes obtained through Sure PetCare API and include
     - pet ID (int) provided by Sure PetCare
     - pet type default always "pet"
     - pet data: raw output from Sure PetCare API
     - pet name (string) name of pet
     - pet state (integer in API, string in script), one of "outside" or "home"

    """

    def __init__(self, data: dict[str, Any]):

        super().__init__(data=data)

        self.pet_id: int = int(data["id"])

        self._type: EntityType = EntityType.PET
        self._data: dict[str, Any] = data

        self._name = str(name) if (name := self._data.get("name")) else "Unnamed"

        self.state = PetState(data["status"]) if "status" in data else "Unknown"

    @property
    def id(self) -> int:
        """ID of the household the pet belongs to."""
        return self.pet_id

    @property
    def tag_id(self) -> int | None:
        """ID of the household the pet belongs to."""
        return int(tag_id) if (tag_id := self._data.get("tag_id")) else None

    @property
    def food_type(self) -> str | None:
        """Type of food."""
        # pylint: disable=used-before-assignment
        return str(FoodType(type_id)) if (type_id := self._data.get("food_type_id")) else None

    @property
    def updated_at(self) -> datetime | None:
        """Type of food."""
        return (
            datetime.fromisoformat(updated_at)
            if (updated_at := self._data.get("updated_at"))
            else None
        )

    @property
    def photo_url(self) -> str | None:
        """Picture of the Pet."""
        picture_url = (
            photo_url
            if (photo_url := self._data.get("photo", {}).get("location"))
            else "https://surehub.io/assets/images/no-pet-pic-dark.svg"
        )

        return urlparse(picture_url).geturl()

    @property
    def at_home(self) -> bool:
        """Location of the Pet."""
        return bool(self.location.where == Location.INSIDE)

    @property
    def location(self) -> PetLocation:
        """Location of the Pet."""
        position = self._data.get("position", {})
        # pylint: disable=no-member
        return PetLocation(
            where=Location(position.get("where", Location.UNKNOWN.value)),
            since=position.get("since", None),
        )

    @property
    def activity(self) -> PetActivity:
        """Last Activity of the Pet."""
        activity = self._data.get("status", {}).get("activity", {})
        # pylint: disable=no-member
        return PetActivity(
            where=Location(activity.get("where", Location.UNKNOWN.value)),
            since=activity.get("since", None),
        )

    @property
    def feeding(self) -> StateFeeding | None:
        """Last Activity of the Pet."""
        if activity := self._data.get("status", {}).get("feeding", {}):
            return StateFeeding(
                change=activity.get("change", [0.0, 0.0]),
                at=datetime.fromisoformat(activity.get("at", None)),
            )

        return None

    @property
    def drinking(self) -> StateDrinking | None:
        """Last Activity of the Pet."""
        if activity := self._data.get("status", {}).get("drinking", {}):
            return StateDrinking(
                change=activity.get("change", [0.0]),
                at=datetime.fromisoformat(activity.get("at", None)),
            )

        return None

    @property
    def last_lunch(self) -> datetime | None:
        return self.feeding.at if self.feeding else None

    @property
    def last_drink(self) -> datetime | None:
        return self.drinking.at if self.drinking else None



"""
surepy
====================================
The core module of surepy

|license-info|
"""

from __future__ import annotations

import asyncio
import logging

from datetime import datetime, time
from http import HTTPStatus
from http.client import HTTPException
from logging import Logger
from os import environ
from pathlib import Path
from typing import Any
from uuid import uuid1

import aiohttp
import async_timeout

from .const import (
    ACCEPT,
    ACCEPT_ENCODING,
    ACCEPT_LANGUAGE,
    API_TIMEOUT,
    AUTH_RESOURCE,
    AUTHORIZATION,
    BASE_RESOURCE,
    CONNECTION,
    CONTENT_TYPE_JSON,
    CONTENT_TYPE_TEXT_PLAIN,
    CONTROL_RESOURCE,
    ETAG,
    HOST,
    HTTP_HEADER_X_REQUESTED_WITH,
    ORIGIN,
    PET_RESOURCE,
    POSITION_RESOURCE,
    REFERER,
    SUREPY_USER_AGENT,
    USER_AGENT,
    DEVICE_TAG_RESOURCE,
)
from .enums import Location, LockState
from .exceptions import SurePetcareAuthenticationError, SurePetcareConnectionError, SurePetcareError


TOKEN_ENV = "SUREPY_TOKEN"  # nosec
TOKEN_FILE = Path("~/.surepy.token").expanduser()

# get a logger
logger: Logger = logging.getLogger(__name__)


def token_seems_valid(token: str) -> bool:
    """check validity of an api token based on its characters and length

    Args:
        token (str): sure petcare api token

    Returns:
        bool: True if ``token`` seems valid
    """
    return (
        (token is not None) and token.isascii() and token.isprintable() and (320 < len(token) < 448)
    )


def find_token() -> str | None:
    token: str | None = None

    # check env token
    if (env_token := environ.get(TOKEN_ENV, None)) and token_seems_valid(token=env_token):
        token = env_token

    # check file token
    elif (
        TOKEN_FILE.exists()
        and (file_token := TOKEN_FILE.read_text(encoding="utf-8"))
        and token_seems_valid(token=file_token)
    ):
        token = file_token

    return token


class SureAPIClient:
    """Communication with the Sure Petcare API."""

    def __init__(
        self,
        email: str | None = None,
        password: str | None = None,
        # loop: Optional[asyncio.AbstractEventLoop] = None,
        auth_token: str | None = None,
        api_timeout: int = API_TIMEOUT,
        session: aiohttp.ClientSession | None = None,
        surepy_version: str | None = None,
    ) -> None:
        """Initialize the connection to the Sure Petcare API."""

        self._session = session

        # sure petcare credentials
        self.email = email
        self.password = password
        # random device id
        self._device_id: str = str(uuid1())

        # connection settings
        self._api_timeout: int = api_timeout

        self._surepy_version: str | None = surepy_version

        # api token management
        self._auth_token: str | None = None
        if auth_token and token_seems_valid(auth_token):
            self._auth_token = auth_token
        elif token := find_token():
            self._auth_token = token
        else:
            # no valid credentials/token
            SurePetcareAuthenticationError("sorry 🐾 no valid credentials/token found ¯\\_(ツ)_/¯")

        # storage for received api data
        self.resources: dict[str, Any] = {}
        # storage for etags
        self._etags: dict[str, str] = {}

        logger.debug("initialization completed | vars(): %s", vars())

    def _generate_headers(self) -> dict[str, str]:
        """Build a HTTP header accepted by the API"""
        user_agent = (
            SUREPY_USER_AGENT.format(version=self._surepy_version) if self._surepy_version else None
        )

        return {
            HOST: "app.api.surehub.io",
            CONNECTION: "keep-alive",
            ACCEPT: f"{CONTENT_TYPE_JSON}, {CONTENT_TYPE_TEXT_PLAIN}, */*",
            ORIGIN: "https://surepetcare.io",
            USER_AGENT: user_agent if user_agent else SUREPY_USER_AGENT,
            REFERER: "https://surepetcare.io",
            ACCEPT_ENCODING: "gzip, deflate",
            ACCEPT_LANGUAGE: "en-US,en-GB;q=0.9",
            HTTP_HEADER_X_REQUESTED_WITH: "com.sureflap.surepetcare",
            AUTHORIZATION: f"Bearer {self._auth_token}",
            "X-Device-Id": self._device_id,
        }

    async def get_token(self) -> str | None:
        """Get or refresh the authentication token."""
        authentication_data: dict[str, str | None] = dict(
            email_address=self.email, password=self.password, device_id=self._device_id
        )

        token: str | None = None

        session = self._session if self._session else aiohttp.ClientSession()

        try:
            raw_response: aiohttp.ClientResponse = await session.post(
                url=AUTH_RESOURCE, data=authentication_data, headers=self._generate_headers()
            )

            if raw_response.status == HTTPStatus.OK:

                response: dict[str, Any] = await raw_response.json()

                if "data" in response and "token" in response["data"]:
                    token = self._auth_token = response["data"]["token"]

            elif raw_response.status == HTTPStatus.NOT_MODIFIED:
                # Etag header matched, no new data available
                pass

            elif raw_response.status == HTTPStatus.UNAUTHORIZED:
                self._auth_token = None
                raise SurePetcareAuthenticationError()

            else:
                logger.debug("Response from %s: %s", AUTH_RESOURCE, raw_response)
                raise SurePetcareError()

            return token

        except asyncio.TimeoutError as error:
            logger.debug("Timeout while calling %s: %s", AUTH_RESOURCE, error)
            raise SurePetcareConnectionError() from error
        except (aiohttp.ClientError, AttributeError) as error:
            logger.debug("Failed to fetch %s: %s", AUTH_RESOURCE, error)
            raise SurePetcareError() from error
        finally:
            if not self._session:
                await session.close()

    async def call(
        self,
        method: str,
        resource: str,
        data: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
        second_try: bool = False,
        **_: Any,
    ) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""

        # logger.debug("")
        # logger.debug("🐾 %s call to: %s", method, resource)
        # if data:
        #     logger.debug("🐾   with data: %s", data)

        if not self._auth_token:
            self._auth_token = await self.get_token()

        if method not in ["GET", "PUT", "POST", "DELETE"]:
            raise HTTPException(f"unknown http method: {method}")

        response_data = None

        session = self._session if self._session else aiohttp.ClientSession()

        try:
            with async_timeout.timeout(self._api_timeout):
                headers = self._generate_headers()

                # use etag if available
                if resource in self._etags:
                    headers[ETAG] = str(self._etags.get(resource))
                    # logger.debug("🐾 \x1b[38;2;255;26;102m·\x1b[0m etag: %s", headers[ETAG])

                await session.options(resource, headers=headers)
                response: aiohttp.ClientResponse = await session.request(
                    method, resource, headers=headers, data=data, json=json
                )

                if response.status == HTTPStatus.OK or response.status == HTTPStatus.CREATED:

                    self.resources[resource] = response_data = await response.json()

                    if ETAG in response.headers:
                        self._etags[resource] = response.headers[ETAG].strip('"')

                elif response.status == HTTPStatus.NOT_MODIFIED:
                    # Etag header matched, no new data available
                    logger.debug(
                        "🐾 \x1b[38;2;0;255;0m·\x1b[0m %d: etag matched - no new data available",
                        response.status,
                    )

                elif response.status == HTTPStatus.UNAUTHORIZED:
                    logger.error(
                        "🐾 \x1b[38;2;255;26;102m·\x1b[0m %s %s: %d | %s",
                        method,
                        resource.replace("https://", ""),
                        response.status,
                        response,
                    )
                    self._auth_token = None
                    if not second_try:
                        token_refreshed = await self.get_token()
                        if token_refreshed:
                            await self.call(method="GET", resource=resource, second_try=True)

                    raise SurePetcareAuthenticationError()

                else:
                    logger.info(
                        "🐾 \x1b[38;2;255;0;255m·\x1b[0m %s %s: %d | %s",
                        method,
                        resource.replace("https://", ""),
                        response.status,
                        response,
                    )

                if response_data:
                    responselen = len(response_data.get("data", 0))
                else:
                    responselen = 0
                logger.debug(
                    "🐾 \x1b[38;2;0;255;0m·\x1b[0m %s %s | %d",
                    method,
                    resource.replace("https://", ""),
                    responselen,
                )

                if method == "DELETE" and response.status == HTTPStatus.NO_CONTENT:
                    # TODO: this does not return any data, is there a better way?
                    return "DELETE 204 No Content"

                return response_data

        except (asyncio.TimeoutError, aiohttp.ClientError) as error:
            logger.error("Can not load data from %s", resource)
            raise SurePetcareConnectionError() from error
        finally:
            if not self._session:
                await session.close()

    async def get_pets(self) -> list[dict[str, Any]] | None:
        """Retrieve the pet data/state."""
        resource = PET_RESOURCE

        response_data: list[dict[str, Any]] | None = []

        response: dict[str, Any] | None = await self.call(method="GET", resource=resource)
        if response:
            response_data = response.get("data")

        return response_data

    async def set_pet_location(self, pet_id: int, location: Location) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        resource = POSITION_RESOURCE.format(BASE_RESOURCE=BASE_RESOURCE, pet_id=pet_id)

        if location not in [Location.INSIDE, Location.OUTSIDE]:
            raise ValueError(f"Unknown location: {location.name.title()}")

        data = {
            "where": int(location.value),
            "since": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
        }

        if (response := await self.call(method="POST", resource=resource, data=data)) and (
            response_data := response.get("data")
        ):

            desired_state = data.get("where")
            state = response_data.get("where")

            # check if the state is correctly updated
            if state == desired_state:
                return response

        raise SurePetcareError(f"Setting position of {pet_id} failed!")

    async def lock(self, device_id: int) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        return await self._set_lock_state(device_id, LockState.LOCKED_ALL)

    async def lock_in(self, device_id: int) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        return await self._set_lock_state(device_id, LockState.LOCKED_IN)

    async def lock_out(self, device_id: int) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        return await self._set_lock_state(device_id, LockState.LOCKED_OUT)

    async def unlock(self, device_id: int) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        return await self._set_lock_state(device_id, LockState.UNLOCKED)

    async def _set_lock_state(self, device_id: int, mode: LockState) -> dict[str, Any] | None:
        """Retrieve the flap data/state."""
        resource = CONTROL_RESOURCE.format(BASE_RESOURCE=BASE_RESOURCE, device_id=device_id)

        data = {"locking": int(mode.value)}

        if (
            response := await self.call(
                method="PUT", resource=resource, device_id=device_id, data=data
            )
        ) and (response_data := response.get("data")):

            desired_state = data.get("locking")
            state = response_data.get("locking")

            # check if the state is correctly updated
            if state == desired_state:
                return response

        # return None
        raise SurePetcareError("ERROR (UN)LOCKING DEVICE - PLEASE CHECK IMMEDIATELY!")

    async def set_curfew(
        self, device_id: int, lock_time: time, unlock_time: time
    ) -> dict[str, Any] | None:
        """Set the flap curfew times, using the household's timezone"""

        resource = CONTROL_RESOURCE.format(BASE_RESOURCE=BASE_RESOURCE, device_id=device_id)

        data = {
            "curfew": [
                {
                    "lock_time": lock_time.strftime("%H:%M"),
                    "unlock_time": unlock_time.strftime("%H:%M"),
                    "enabled": True,
                }
            ]
        }

        if (
            response := await self.call(
                method="PUT", resource=resource, device_id=device_id, json=data
            )
        ) and (response_data := response.get("data")):

            desired_state = data.get("curfew")
            state = response_data.get("curfew")

            # check if the state is correctly updated
            if state == desired_state:
                return response

        # return None
        raise SurePetcareError("ERROR SETTING CURFEW - PLEASE CHECK IMMEDIATELY!")

    async def _add_tag_to_device(self, device_id: int, tag_id: int) -> dict[str, Any] | None:
        """Add the specified tag ID to the specified device ID"""
        resource = DEVICE_TAG_RESOURCE.format(BASE_RESOURCE=BASE_RESOURCE, device_id=device_id, tag_id=tag_id)

        if(
            response := await self.call(
                method="PUT", resource=resource
            )
        ):
            return response

    async def _remove_tag_from_device(self, device_id: int, tag_id: int) -> dict[str, Any] | None:
        """Removes the specified tag ID from the specified device ID"""
        resource = DEVICE_TAG_RESOURCE.format(BASE_RESOURCE=BASE_RESOURCE, device_id=device_id, tag_id=tag_id)

        if(
            response := await self.call(
                method="DELETE", resource=resource
            )
        ):
            return response



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from pprint import pformat
from typing import Any

from surepy.enums import EntityType, Location


class SurepyEntity(ABC):
    def __init__(self, data: dict[str, Any]):

        # sure petcare id
        self._id: int = int(data.get("id", data.get("_id")))

        # self._sac: SureAPIClient = sac
        self._data = data
        self._type = EntityType(int(data.get("product_id", 0)))

        self._name: str = self._data.get("name", "Unknown")

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"{self._type.name.capitalize()}(data={pformat(self._data)})"

    @property
    def id(self) -> int:
        return self._id

    @property
    def unique_id(self) -> str:
        return f"{self.household_id}-{self.id}"

    @property
    def name(self) -> str:
        return self._name

    @property
    def full_name(self) -> str:
        return f"{self._type.name}_{self.name}"

    @property
    def type(self) -> EntityType:
        return self._type

    @property
    def household_id(self) -> int:
        """ID of the household the entity belongs to."""
        return int(self._data["household_id"])

    def raw_data(self) -> dict[str, Any]:
        return self._data


@dataclass
class StateFeeding:
    change: list[float]
    at: datetime | None


@dataclass
class StateDrinking:
    change: list[float]
    at: datetime | None


@dataclass
class PetLocationData:

    where: Location
    since: datetime | None

    def __str__(self) -> str:
        return self.where.name.title()


@dataclass
class PetActivity(PetLocationData):
    pass


@dataclass
class PetLocation(PetLocationData):
    pass
